#if defined(_MAKROSA)
#include "stdafx.h"
#include "makrosa.h"
#include "global.h"
#include "utilskdlg.h"
#else
#endif
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include "interface.h"
#include "ladedatei.h"
#include "getdata.h"

int geomonly;	// only geometry data
int wnastran;  // > 1: NASTRAN file will be written

int kznode;		// > 0, if node id != index
int kzelem;		// > 0, if element id != index
#define IXGR 3	/* index gruppe */
#define IXMTYP 2  /* index mechtyp */

void globalforce(int ks,float pkt[2][3],float knlast[3]);

int getStructData(char *project,StructData *structdata,int global)
{
	static int etype[21][3] = 
		{1,1,1, 20,2,5,  22,2,5,  30,3,3,  32,3,6,  33,3,9,
		37,3,10,  40,4,4,  42,4,8,  43,4,12,  46,4,9,  47,4,16,
		60,7,6,   62,7,15, 63,7,24, 70,5,4,   72,5,10, 73,5,16,
		80,8,8,   82,8,20, 83,8,32};
	int netype = sizeof(etype)/sizeof(etype[0]);
	memset(structdata,0,sizeof(StructData));

	char title[265];
	char *pchr;

	typedef struct
	{
		int max,tmp[10];
		union
		{
			int irks[1][7][3];
			float rks[1][7][3];
		} rks;
	} coorsys;

	strcpy(title,project);
	if((pchr = strrchr(title,'.')) && (strlen(pchr) <= 4)) *pchr = '\0';
	strcat(title,".fes");

	int param[12];
	memset(param,0,12*sizeof(int));
	int *tmp = NULL;

	int rtw = getFEStruct(3,title,param,tmp,(float*)tmp,
		tmp,tmp,tmp,tmp,tmp,(char *)tmp,(float*)tmp,tmp);

	int nev = param[1];
	int nn1 = param[2];
	int npv = param[0];
	int nln = param[6];
	int ncoors = param[3];
	if(rtw < 0 || nev <= 0 || npv <= 0) return 1;

	int *nrp = new int[npv+1];
	float (*pkt)[3] = new float[npv+1][3];
	int **nel = (int **) new char[(nev+1)*sizeof(int *)];
	nel[0] = new int[nn1*(nev+1)*sizeof(int)];
	for(int k=1;k<=nev;k++)
		nel[k] = nel[k-1] + nn1;
	int ltemp = 4*(nev+1)+100;
	if(nln > ltemp) ltemp = nln/sizeof(int) + 10;
	tmp = new int[ltemp];
	float *ftmp = new float[(ncoors+1)*21];
	coorsys *koor = (coorsys *) ftmp;

	rtw = getFEStruct(2,title,param,nrp,(float *)pkt,nel[0],tmp,
		tmp,tmp,tmp,(char *)tmp,ftmp,tmp);

	if(rtw < 0)
	{
		delete [] nrp;
		delete [] pkt;
		delete [] nel[0];
		delete [] nel;
		delete [] tmp;
		delete [] ftmp;
		return 1;
	}
	structdata->nodedata = new NodeData[npv];
	structdata->elemdata = new ElementData[nev];
	memset(structdata->nodedata,0,npv*sizeof(NodeData));
	memset(structdata->elemdata,0,nev*sizeof(ElementData));

	kznode = 0;
	for(k=1;k<=npv;k++)
	{
		if(nrp[k] != k) kznode = 1;
		structdata->nodedata[k-1].id = nrp[k];
		memcpy(structdata->nodedata[k-1].xyz,pkt[k],3*sizeof(float));
		structdata->nodedata[k-1].ndf = 6;
		structdata->nodedata[k-1].gtype = 'G';
		structdata->nodedata[k-1].pspc = 0;
	}

	kzelem = 0;
	int l,la = 0;
	for(k=1;k<=nev;k++)
	{
		int gt = nel[k][1];
		l = la;
		if(etype[l][0] != gt)
		{
			for(l=0;l<netype;l++)
				if(etype[l][0] == gt) break;
			if(l == netype) continue;
			la = l;
		}
		int nodes = etype[l][2];
		if(nel[k][0] != k) kzelem = 1;
		structdata->elemdata[k-1].id = nel[k][0];
		structdata->elemdata[k-1].shape = etype[l][1];
		structdata->elemdata[k-1].group = nel[k][IXGR];
		structdata->elemdata[k-1].mechtype = nel[k][IXMTYP];
		structdata->elemdata[k-1].nodes = nodes;
		memcpy(structdata->elemdata[k-1].lnodes,&nel[k][4],nodes*sizeof(int));
	}
	int ix = 0;
	for(int kc=0;kc<ncoors;kc++)
	{
		int kt = koor->rks.irks[kc][0][1];
		if(kt <= 3) ++ix;
	}
	ncoors = ix;
	ix = 0;
	if(ncoors)
	{
		structdata->coorsys = new CoorSystem[ncoors];
		memset(structdata->coorsys,0,ncoors*sizeof(CoorSystem));
		for(int kc=0;kc<ncoors;kc++)
		{
			int kt = koor->rks.irks[kc][0][1];
			if(kt > 3) continue;
			structdata->coorsys[ix].id = koor->rks.irks[kc][0][0];
			structdata->coorsys[ix].ktyp = kt;
			float z[3],xz[3];
			for(int i=0;i<3;i++)
			{
				z[i] = koor->rks.rks[kc][1][i] + koor->rks.rks[kc][6][i];
				xz[i] = koor->rks.rks[kc][1][i] + koor->rks.rks[kc][6][i] + koor->rks.rks[kc][4][i];
			}
			memcpy(structdata->coorsys[ix].coors[0],koor->rks.rks[kc][1],3*sizeof(float));
			memcpy(structdata->coorsys[ix].coors[1],z,3*sizeof(float));
			memcpy(structdata->coorsys[ix].coors[2],xz,3*sizeof(float));
			++ix;
		}
	}
	structdata->ne = nev;
	structdata->np = npv;
	structdata->ncoor = ix;
	delete [] nrp;
	delete [] nel[0];
	delete [] nel;
	delete [] tmp;
	delete [] ftmp;
	delete [] pkt;

	if(geomonly) return 0;

	getLoadData(project,structdata,global);
	return 0;
}

int getLoadData(char *project,StructData *structdata,int global)
{
	static int surf80[6][4] = {4,3,2,1, 1,2,6,5, 2,3,7,6, 3,4,8,7, 4,1,5,8, 5,6,7,8};
	static int surf60[5][4] = {3,2,1,0, 1,2,5,4, 2,3,6,5, 3,1,4,6, 4,5,6,0};
	static int surf70[4][4] = {3,2,1,0, 1,2,4,0, 2,3,4,0, 3,1,4,0};
	int (*surfx)[4] = NULL;

	long *kn_muster = NULL;
	int *knoten = NULL;
	float *knlast = NULL;
	long *el_muster = NULL;
	int *elnr = NULL;
	float *el_last = NULL;
	long *kr_muster = NULL;
	int *kr_kn = NULL;
	double **material = NULL;
	long *qs_muster = NULL;
	int *knqs = NULL;
	float *qsww = NULL;
	float (*kkr)[4] = NULL;

	int param[30];
	structdata->nm = 0;
	structdata->nelp = 0;
	structdata->nel = 0;

	NodeData *pnode = structdata->nodedata;
	int mnodes = structdata->np;
	ElementData *pelem = structdata->elemdata;
	int nelem = structdata->ne;

	char title[265];
	char *pchr;
	strcpy(title,project);
	if((pchr = strrchr(title,'.')) && (strlen(pchr) <= 4)) *pchr = '\0';
	strcat(title,".lqd");

	int kf = 3;
	memset(param,0,30*sizeof(int));
	int rtw = getLoaddata(kf,title,param,kn_muster,knoten,knlast,el_muster,
		elnr,el_last,kr_muster,kr_kn,material,qs_muster,knqs,qsww,(float *)kkr);

	if(rtw < 0) return -1;

	int kn_nmuster = param[0];
	int nkn = param[1];
	int kn_lmuster = param[2];
	int kn_llast = param[3];
	int el_nmuster = param[4];
	int nelnr = param[5];
	int el_lmuster = param[6];
	int el_llast = param[7];
	int kr_nmuster = param[8];
	int nkrkn = param[9];
	int kr_lmuster = param[12];
	int n_material = param[13];
	int qs_nmuster = param[14];
	int nknqs = param[15];
	int qs_lmuster = param[16];
	int nwqs = param[17];
	int nkkr = param[18];

	if(kn_nmuster > 0)
	{
		kn_muster = new long[kn_lmuster*(kn_nmuster+1)*sizeof(long)];
		knoten = new int[nkn*sizeof(int)];
		knlast = new float[kn_llast*kn_nmuster*sizeof(float)];
	}
	if(el_nmuster > 0)
	{
		el_muster = new long[el_lmuster*(el_nmuster+1)*sizeof(long)];
		elnr = new int[nelnr*sizeof(int)];
		el_last = new float[el_llast*el_nmuster*sizeof(float)];
		kkr = new float[nkkr+10][4];
	}
	if(kr_nmuster > 0)
	{
		kr_muster = new long[kr_lmuster*(kr_nmuster+1)*sizeof(long)];
		kr_kn = new int[nkrkn*sizeof(int)];
	}
	if(n_material)
	{
		material = (double **) new char[n_material*sizeof(double *)];
		material[0] = new double[n_material*110*sizeof(double)];
		for(int k=1;k<n_material;k++)
		{
			material[k] = material[k-1] + 110;
		}
	}
	if(qs_nmuster)
	{
		qs_muster = new long[qs_lmuster*(qs_nmuster+1)*sizeof(long)];
		knqs = new int[nknqs*sizeof(int)];
		qsww = new float[nwqs*sizeof(float)];
	}

	kf = 2;
	rtw = getLoaddata(kf,title,param,kn_muster,knoten,knlast,el_muster,
		elnr,el_last,kr_muster,kr_kn,material,qs_muster,knqs,qsww,(float *)kkr);

	if(rtw < 0) return -1;

	void *kxxn = NULL;
	void *kxxe = NULL;
	int *nre = NULL;
	int *nrp = NULL;
	int ixn,mmnodes = 0;
	int ixe,mmelem = 0;
	if(kznode)
	{
		nrp = new int[mnodes+1];
		for(int k=0;k<mnodes;k++)
			nrp[k+1] = pnode[k].id;
	}
	if(kzelem)
	{
		nre = new int[nelem+1];
		for(int k=0;k<nelem;k++)
			nre[k+1] = pelem[k].id;
	}

	int ks,id,ix,ixx,nf=0;
	if(kn_nmuster > 0 || nkkr > 0) // node forces
	{
		int kzg=0;
		structdata->nodeforces = new NodeForces[nkn+nkkr+1];
		float fdata[3];
		for(int k=1;k<=kn_nmuster;k++)
		{
			int lsid = (int) kn_muster[(k-1)*kn_lmuster+0];
			int ksg = (int) kn_muster[(k-1)*kn_lmuster+2];
			int ksi = (int) kn_muster[(k-1)*kn_lmuster+3];
			int type = (int) kn_muster[(k-1)*kn_lmuster+5];
			int nk = (int) kn_muster[(k-1)*kn_lmuster+1];
			ixx = (int) kn_muster[(k-1)*kn_lmuster+4];

			if(kznode) // node id to index
			{
				suche_index(&knoten[ixx],nk,nrp,mnodes,&kxxn);
			}

			memset(fdata,0,3*sizeof(float));
			ix = kn_llast*(k-1);
			memcpy(fdata,&knlast[ix],3*sizeof(float));
			float scale = knlast[ix+3];
			for(int l=0;l<nk;l++)
			{
				ks = ksg;
				ixn = knoten[ixx+l]-1;
				if(ixn < 0 || ixn >= mnodes)
				{
					++mmnodes;
					continue;
				}
				id = pnode[ixn].id;
				if(ksi) ks = id + ksi;
				// transform nodeforces to global coordinates
#if defined(_MAKROSA)
				if(ks && global)
				{
					if(ixn >= mnodes) continue;
					float pkt[2][3],vv[6];
					memcpy(vv,&knlast[(k-1)*kn_llast],3*sizeof(float));
					memcpy(pkt,structdata->nodedata[ixn].xyz,3*sizeof(float));
					globalforce(ks,pkt,vv);
					ks = 0;
					memcpy(fdata,vv,3*sizeof(float));;
				}
#endif
				structdata->nodeforces[nf].id = id;
				structdata->nodeforces[nf].loadset = lsid;
				structdata->nodeforces[nf].cid = ks;
				structdata->nodeforces[nf].type = type;
				structdata->nodeforces[nf].scale = scale;
				memcpy(structdata->nodeforces[nf].fdata,fdata,3*sizeof(float));
				++nf;
			}
		}
		delete [] kn_muster;
		delete [] knoten;
		delete [] knlast;

		// node forces from distributed loads
		if(nkkr > 0)
		{
			for(int k=1;k<=el_nmuster;k++)
			{
				int typ = (int) el_muster[(k-1)*el_lmuster+11];
				if(!typ) continue;
				int lsid = (int) el_muster[(k-1)*el_lmuster+0];
				int nk = (int) el_muster[(k-1)*el_lmuster+12];
				int ixkkr = (int) el_muster[(k-1)*el_lmuster+14];
				int *ixnode = new int[nk+1];
				for(int ik=0;ik<nk;ik++)
					ixnode[ik] = (int) kkr[ixkkr+ik][0];

				if(kznode) // node id to index
				{
					suche_index(ixnode,nk,nrp,mnodes,&kxxn);
				}
				for(ik=0;ik<nk;ik++)
				{
					ixn = ixnode[ik]-1;
					if(ixn < 0 || ixn >= mnodes)
					{
						++mmnodes;
						continue;
					}
					id = pnode[ixn].id;
					memcpy(fdata,&kkr[ixkkr+ik][1],3*sizeof(float));
					structdata->nodeforces[nf].scale = 1.f;
					structdata->nodeforces[nf].id = id;
					structdata->nodeforces[nf].type = 0;
					structdata->nodeforces[nf].loadset = (int) lsid;
					structdata->nodeforces[nf].cid = 0;
					memcpy(structdata->nodeforces[nf].fdata,fdata,3*sizeof(float));
					++nf;
				}
				delete [] ixnode;
			}
		}
		structdata->nf = nf;
	}

	int nel = 0;
	int nelmax = nelnr+1;
	if(el_nmuster > 0) // distributed loads
	{
		structdata->elemloads = new DistributedLoads[nelmax];
		for(int k=1;k<=el_nmuster;k++)
		{
			int typ = (int) el_muster[(k-1)*el_lmuster+11];
			if(typ > 0) continue;
			int nastrantyp = (int) el_muster[(k-1)*el_lmuster+15];
			if(wnastran && (nastrantyp < 10 || nastrantyp > 14)) continue;
			int lsid = (int) el_muster[(k-1)*el_lmuster+0];
			int nrkant = (int) el_muster[(k-1)*el_lmuster+1];
			int nrobf = (int) el_muster[(k-1)*el_lmuster+2];
			int icomp = (int) el_muster[(k-1)*el_lmuster+3];
			int eflag = (int) el_muster[(k-1)*el_lmuster+4];
			int node = (int) el_muster[(k-1)*el_lmuster+5];
			int ndw1 = (int) el_muster[(k-1)*el_lmuster+6];
			int ndw2 = (int) el_muster[(k-1)*el_lmuster+7];
			int gflag = (int) el_muster[(k-1)*el_lmuster+8];
			int ltyp = (int) el_muster[(k-1)*el_lmuster+9];
			if(ltyp < 0 || ltyp > 1) ltyp = 1;
			if(wnastran) ltyp = 1;
			int kwahl = (int) el_muster[(k-1)*el_lmuster+16];
			int ne = (int) el_muster[(k-1)*el_lmuster+10];
			int ixx = (int) el_muster[(k-1)*el_lmuster+13];
			int nfe = nrobf;
			if(!ltyp) nfe = nrkant;
			int ndw = ndw1+ndw2;
			if(ndw > 32) ndw = 32;
			float pdata[32];
			memset(pdata,0,32*sizeof(float));
			memcpy(pdata,&el_last[(k-1)*el_llast],ndw*sizeof(float));

			int (*idelem)[6] = NULL;
			if(!kwahl)
			{
				idelem = new int[nelem][6];
			}
			if(kwahl)
			{
				delete [] idelem;
				// Node selection given: idelem[][6] = ie,ido,k1,k2,k3,k4
				idelem = new int[6*nelem][6];
				int *kzn = new int[mnodes+1];
				memset(kzn,0,(mnodes+1)*sizeof(int));
				if(kznode) // node id to index
				{
					suche_index(&elnr[ixx],ne,nrp,mnodes,&kxxn);
				}
				for(int l=0;l<ne;l++)
				{
					ixn = elnr[ixx+l];
					if(ixn > 0 && ixn <= mnodes)
						kzn[ixn] = 1;
					else
						++mmnodes;
				}
				ne = 0;
				for(int ie=0;ie<nelem;ie++)
				{
					int shape = pelem[ie].shape;
					int nk = shape;
					if(shape >= 2 && shape <= 4)
					{
						for(int ik=0;ik<nk;ik++)
						{
							int nix = pelem[ie].lnodes[ik];
							if(!kzn[nix]) break;
						}
						if(ik == nk) // all nodes are in the selection
						{
							memset(idelem[ne],0,6*sizeof(int));
							idelem[ne][0] = ie;
							memcpy(&idelem[ne][2],pelem[ie].lnodes,nk*sizeof(int));
							++ne;
						}
					} else if(shape > 4 && shape <= 8) // solid element
					{
						int no;
						if(shape == 8) // hex-element
						{ 
							surfx = surf80;
							no = 6;
						} else if(shape == 7) // wedge
						{
							surfx = surf60;
							no = 5;
						} else if(shape == 5) // tet
						{
							surfx = surf70;
							no = 4;
						} else
							continue;

						for(int io=0;io<no;io++)
						{
							int mk = 4;
							if(!surfx[io][3]) mk = 3;
							for(int i=0;i<mk;i++)
							{
								int ix = surfx[io][i];
								ix = pelem[ie].lnodes[ix-1];
								if(!kzn[ix]) break;
							}
							if(i < mk) continue;

							memset(idelem[ne],0,6*sizeof(int));
							idelem[ne][0] = ie;
							idelem[ne][1] = io + 1;
							for(i=0;i<mk;i++)
							{
								int ix = surfx[io][i];
								ix = pelem[ie].lnodes[ix-1];
								idelem[ne][i+2] = ix;
							}
							++ne;
						}
					}
				}
			} else // elementselection
			{
				if(kzelem) // element id to index
				{
					suche_index(&elnr[ixx],ne,nre,nelem,&kxxe);
				}
				int nne = ne;
				ne = 0;
				for(int l=0;l<nne;l++)
				{
					ixe = elnr[ixx+l]-1;
					if(ixe < 0 || ixe >= nelem)
					{
						++mmelem;
						continue;
					}
					int nk = pelem[ixe].shape;
					memset(idelem[ne],0,6*sizeof(int));
					idelem[ne][0] = ixe;
					memcpy(&idelem[ne][2],pelem[ixe].lnodes,nk*sizeof(int));
					++ne;
				}
			}

			if(wnastran && (nastrantyp == 11 || nastrantyp == 12))
			{
				int mmm = 30;
				int ix0 = 1;
				int mnn = 1;
				if(nastrantyp == 11)
				{
					mmm = 25;
					ix0 = 6;
					mnn = 6;
				}
				int mm = (ne-1)/mmm+1;
				int m = mmm;
				if(m > ne) m = ne;
				int ixxx = ixx;
				if(nel+mm > nelmax-1)
				{
					DistributedLoads *pd = structdata->elemloads;
					nelmax += 10*mm;
					structdata->elemloads = new DistributedLoads[nelmax];
					memcpy(structdata->elemloads,pd,nel*sizeof(DistributedLoads));
					delete [] pd;
				}
				for(int l=0;l<mm;l++)
				{
					structdata->elemloads[nel].loadset = lsid;
					structdata->elemloads[nel].ltyp = ltyp;
					structdata->elemloads[nel].nastrantyp = nastrantyp;
					structdata->elemloads[nel].eflag = eflag;
					structdata->elemloads[nel].gflag = gflag;
					structdata->elemloads[nel].icomp = icomp;
					structdata->elemloads[nel].node = node;
					structdata->elemloads[nel].nfe = nfe;
					memset(structdata->elemloads[nel].pdata,0,32*sizeof(float));
					memcpy(structdata->elemloads[nel].pdata,pdata,mnn*sizeof(float));
					for(int i=0;i<m;i++)
						structdata->elemloads[nel].pdata[ix0+i] = (float) elnr[ixxx+i];
					ixxx += mmm;
					ne -= mmm;
					if(m > ne) m = ne;
					nel++;
				}
			} else
			{
				if(nel+ne > nelmax-1)
				{
					DistributedLoads *pd = structdata->elemloads;
					nelmax += ne+10;
					structdata->elemloads = new DistributedLoads[nelmax];
					memcpy(structdata->elemloads,pd,nel*sizeof(DistributedLoads));
					delete [] pd;
				}
				for(int l=0;l<ne;l++)
				{
					int ixe = idelem[l][0];
					int ide = pelem[ixe].id;
					if(!kwahl)
					{
						structdata->elemloads[nel].id = elnr[ixx+l];
					}
					else
					{
						structdata->elemloads[nel].id = ide;
						nfe = idelem[l][1];
					}
					structdata->elemloads[nel].loadset = lsid;
					structdata->elemloads[nel].ltyp = ltyp;
					structdata->elemloads[nel].nastrantyp = nastrantyp;
					structdata->elemloads[nel].eflag = eflag;
					structdata->elemloads[nel].gflag = gflag;
					structdata->elemloads[nel].icomp = icomp;
					structdata->elemloads[nel].node = node;
					structdata->elemloads[nel].nfe = nfe;
					structdata->elemloads[nel].ndw = ndw;
					memcpy(structdata->elemloads[nel].pdata,pdata,32*sizeof(float));
					if(wnastran && nastrantyp == 10)
					{
						memset(&structdata->elemloads[nel].pdata[1],0,31*sizeof(float));
						for(int i=0;i<4;i++)
							structdata->elemloads[nel].pdata[i+1] = (float) idelem[l][i+2];
					}
					if(wnastran)
					{
						int g1 = idelem[l][2];
						int g4 = idelem[l][4];
						if(!idelem[l][5]) g4 = 0;
						
						structdata->elemloads[nel].pdata[4] = (float) g1;
						structdata->elemloads[nel].pdata[5] = (float) g4;
					}
					++nel;
				}
			}
			delete [] idelem;
		}
		delete [] el_muster;
		delete [] elnr;
		delete [] el_last;
		delete [] kkr;
		structdata->nel = nel;
	}
	if(kr_nmuster > 0)
	{
		int kzm=0;
		int np = structdata->np;
		char gtype;
		for(int k=1;k<=kr_nmuster;k++)
		{
			int ndf = (int) kr_muster[(k-1)*kr_lmuster+0];
			gtype = (char) kr_muster[(k-1)*kr_lmuster+1];
			int config = (int) kr_muster[(k-1)*kr_lmuster+2];
			int pspc = (int) kr_muster[(k-1)*kr_lmuster+3];
			int cid = (int) kr_muster[(k-1)*kr_lmuster+4];
			int inkk = (int) kr_muster[(k-1)*kr_lmuster+5];
			int nk = (int) kr_muster[(k-1)*kr_lmuster+6];
			ixx = (int) kr_muster[(k-1)*kr_lmuster+7];
			if(kznode) // node id to index
			{
				suche_index(&kr_kn[ixx],nk,nrp,mnodes,&kxxn);
			}
			for(int l=0;l<nk;l++)
			{
				ixn = kr_kn[ixx+l]-1;
				if(ixn < 0 || ixn >= mnodes)
				{
					++mmnodes;
					continue;
				}
				int id = kr_kn[ixn];
				if(inkk) cid = id + inkk;
				structdata->nodedata[ixn].ndf = ndf;
				structdata->nodedata[ixn].config;
				// add flags
				if(structdata->nodedata[ixn].pspc == 0)
				{
					structdata->nodedata[ixn].pspc = pspc;
					structdata->nodedata[ixn].cid = cid;
				} else
				{
					if(structdata->nodedata[ixn].cid == cid)
					{
						char spcd[10],spc[10] = "         ";
						int ix;
						sprintf(spcd,"%d",structdata->nodedata[ixn].pspc);
						for(int i=0;i<strlen(spcd);i++)
						{
							ix = spcd[i] - '0';
							if(ix > 0 && ix <= 6) spc[ix] = '1';
						}
						sprintf(spcd,"%d",pspc);
						for(i=0;i<strlen(spcd);i++)
						{
							ix = spcd[i] - '0';
							if(ix > 0 && ix <= 6) spc[ix] = '1';
						}
						int pspcn = 0;
						for(i=1;i<=6;i++)
							if(spc[i] == '1') pspcn = 10*pspcn + i;
						structdata->nodedata[ixn].pspc = pspcn;
					} else
					{
#if defined(_MAKROSA)
						if(!kzm) achtung(IDS_STRING63586);
#endif
						kzm = 1;
					}
				}
				structdata->nodedata[ixn].gtype = gtype;
			}
		}
		delete [] kr_muster;
		delete [] kr_kn;
	}

	if(n_material)
	{
		structdata->mprop = new MaterialProperties[n_material];
		memset(structdata->mprop,0,n_material*sizeof(MaterialProperties));
		for(int k=0;k<n_material;k++)
		{
			int id = (int) material[k][0];
			int type = (int) material[k][1];
			int nw = (int) material[k][2];
			structdata->mprop[k].id = id;
			structdata->mprop[k].type = type;
			structdata->mprop[k].n1 = 0;
			memcpy(structdata->mprop[k].data,&material[k][10],nw*sizeof(double));;
		}
		structdata->nm = n_material;
		delete [] material[0];
		delete [] material;
	}
	if(qs_nmuster > 0)
	{
		int config = 0;
		int nev = structdata->ne;
		structdata->elemprop = new ElementProperties[qs_nmuster];
		memset(structdata->elemprop,0,qs_nmuster*sizeof(ElementProperties));
		for(int k=0;k<qs_nmuster;k++)
		{
			int pid = (int)qs_muster[k*qs_lmuster];
			structdata->elemprop[k].id = pid;
			structdata->elemprop[k].matid = (int)qs_muster[k*qs_lmuster+1];
			config = structdata->elemprop[k].config = (int)qs_muster[k*qs_lmuster+2];
			int nv = (int)qs_muster[k*qs_lmuster+3];
			if(nv > EPM) nv = EPM;
			structdata->elemprop[k].nv = nv;
			structdata->elemprop[k].shape = (int)qs_muster[k*qs_lmuster+4];
			structdata->elemprop[k].nodes = (int)qs_muster[k*qs_lmuster+5];
			ixx = (int)qs_muster[k*qs_lmuster+8];
			memcpy(structdata->elemprop[k].data,&qsww[ixx],nv*sizeof(float));
			int ne = (int)qs_muster[k*qs_lmuster+6];
			ixx = (int)qs_muster[k*qs_lmuster+7];
			if(kzelem) // element id to index
			{
				suche_index(&knqs[ixx],ne,nre,nelem,&kxxe);
			}
			for(int l=0;l<ne;l++)
			{
				ixe = knqs[ixx+l]-1;
				if(ixe < 0 || ixe >= nelem)
				{
					++mmelem;
					continue;
				}
				int ide = pelem[ixe].id;
				if(wnastran && config < 1000)
				{
					structdata->elemdata[ixe].pid2 = pid;
				} else
					structdata->elemdata[ixe].pid = pid;
			}
		}
		structdata->nelp = qs_nmuster;

		delete [] qs_muster;
		delete [] knqs;
		delete [] qsww;
	}
	delete [] kxxn;
	delete [] kxxe;
	delete [] nre;
	delete [] nrp;
	char txt[200];
	if(mmnodes)
	{
#if defined(_MAKROSA)
		CString ss;
		ss.LoadString(IDS_STRING63641);
		sprintf(txt,ss,mmnodes);
		achtung(txt);
#else
		sprintf(txt,"Die Lastdatei enthlt %d externe Knotennummern,\n die nicht existieren",mmnodes);
#endif
	}
	if(mmelem)
	{
#if defined(_MAKROSA)
		CString ss;
		ss.LoadString(IDS_STRING63640);
		sprintf(txt,ss,mmelem);
		achtung(txt);
#else
		sprintf(txt,"Die Lastdatei enthlt %d externe Elementnummern,\n die nicht existieren",mmnodes);
#endif
	}
	return 0;
}
// free all memory allocated in getStructData (function in interface.lib)
void freeStructData(StructData *structdata)
{
	delete [] structdata->nodedata;
	delete [] structdata->elemdata;
	delete [] structdata->mprop;
	delete [] structdata->elemprop;
	delete [] structdata->coorsys;
	delete [] structdata->elemloads;
	delete [] structdata->nodeforces;
	delete structdata;
}
// seach indices
// ixx[0-nxx] : (ein) id (out) index
// nr[1-nnr] : ids
// if *kxx > NULL: kxx[][2] = ids/indeces sorted
void suche_index(int ixx[],int nxx,int nr[],int nnr,void **kxx)
{
	// id == index ?
	for(int l=0;l<nxx;l++)
	{
		int ix = ixx[l];
		if(ix > nnr || ix < 1) break;
		if(nr[ix] != ix) break;
	}
	if(l == nxx) return;

	int (*ix)[2] = NULL;
	int nix = 0;
	if(!*kxx)
	{
		ix = new int[nnr+2][2];
		nix = 1;
		*kxx = (void *)ix;
		for(int k=1;k<=nnr;k++)
		{
			ix[k][0] = nr[k];
			ix[k][1] = k;
		}
		qsort((void *) &ix[1][0],nnr,2*sizeof(int),intsort);
	} else
		ix = (int (*)[2])*kxx;
	int *p;
	for(int k=0;k<nxx;k++)
	{
		int m = ixx[k];

		p = (int *)bsearch((void *)&m,(void *)&ix[1][0],nnr,2*sizeof(int),intsort);
		if(p)
		{
			ixx[k] = p[1];
		} else
		{
			ixx[k] = -1;
			if(m == 0) ixx[k] = 0;
		}
	}
	if(nix)
	{
		delete [] ix;
		*kxx = NULL;
	}
}
// integer-feld sortieren
int	__cdecl intsort(const void *elem1,const void *elem2)
{
	int m1 = *((int*)elem1);
	int m2 = *((int*)elem2);
	if(m1 < m2)
		return -1;
	else if(m1 > m2)
		return 1;
	return 0;
}
/* kreuzprodukt zweier vektoren */
// kzl = 0/1: auf laenge 1 normieren n/j
// return: vektorlaenge
float a_kreuz_bn(float a[],float b[],float ab[],int kzl)
{
	ab[0] = a[1]*b[2] - a[2]*b[1];
	ab[1] = a[2]*b[0] - a[0]*b[2];
	ab[2] = a[0]*b[1] - a[1]*b[0];

	float s = vlaengen(ab);
	if(kzl && s > 1.e-6)
	{
		for(int i=0;i<3;i++)
			ab[i] /= s;
		s = (float) 1.;
	}
	return s;
}
/* laenge eines vektors */
float vlaengen(float a[])
{
	int i;
	float s = (float) 0.;

	for(i=0;i<3;i++)
		s += a[i]*a[i];

	s = (float) sqrt((double) s);
	return s;
}

