#include "interface.h"
#include "nastran.h"
#include <stdio.h>
#include <time.h>
#include <string.h>
#include <math.h>

void writeNodeDataN(int spc);
void writeElementDataN();
void writeSurfaceElement(ElementData *pe,int ie);
void writeSurfaceElement2(int ixp2,int n,int kzr);
void writeSurfaceElement3(int ixp2,int n);
void writeSurfaceElement4(int ixp2,int n);
void writeSolidElement(ElementData *pe,int ie);
void writeBarElement(ElementData *pe,int ie);
void writeCBarElement(ElementData *pe,int ie);
void writeMaterialPropertiesN();
void writeElementPropertiesN();
void writeCoordinateFramesN();
void writeDistibutedLoadsN();
void writeDistibutedLoadsN10(DistributedLoads *pl);
void writeDistibutedLoadsN11(DistributedLoads *pl);
void writeDistibutedLoadsN12(DistributedLoads *pl);
void writeDistibutedLoadsN14(DistributedLoads *pl);
void writeNodeForcesN();
void writeElement(ElementData *pe);
void write_record(float w[],int n,int nl,int first);
void write_recordm(float w[],int nv,int m1,int first);
void write_recordi(float w[],int n,int nl,int first);
void write_recordg(float w[],int kzi[],int n,int nl,int first);

StructData *structdata;
FILE *pf;
extern int geomonly;	// only geometry data
int *ixprop;  // -> property-record
int *ixprop2; // -> additional element data

int nastran(char *file,StructData *psd,char *project,
	int geom,int mfile,int spc1)
{
	structdata = psd;
	pf = fopen(file,"w");
	if(!pf) return 1;
	int spc = !spc1;
	geomonly = geom;

	if(!geomonly)
	{
		// Title Card
		char *p = strrchr(project,'.');
		if(p && strlen(p) <= 4) *p = '\0';
		fprintf(pf,"$ Project file: %s\n",project);

		struct tm *newtime;
		long ltime;
		time( &ltime );
		newtime = gmtime( &ltime );
		char *date = asctime( newtime);
		fprintf(pf,"$ Date:  %s",date);

		fprintf(pf,"$\nSOL 101\nCEND\n$\n");
		fprintf(pf,"TITLE=%s\n",project);

		fprintf(pf,"ECHO=BOTH\nDISP=ALL\nFORCE=ALL\nSPCF=ALL\nSTRESS=ALL\n");
		
		// load set id
		int nel = psd->nel;
		int nf = psd->nf;
		int idd,ida = 0;
		int mid=0,*id = new int[nel+nf+2];		
		DistributedLoads *elemloads = psd->elemloads;
		for(int l=0;l<nel;l++)
		{
			idd = elemloads[l].loadset;
			if(idd == ida) continue;
			ida = idd;
			for(int i=0;i<mid;i++)
				if(id[i] == idd) break;
			if(i < mid) continue;
			id[mid++] = idd;
		}
		NodeForces *nodeforces = psd->nodeforces;
		for(l=0;l<nf;l++)
		{
			idd = nodeforces[l].loadset;
			if(idd == ida) continue;
			ida = idd;
			for(int i=0;i<mid;i++)
				if(id[i] == idd) break;
			if(i < mid) continue;
			id[mid++] = idd;
		}
		for(l=0;l<mid;l++)
		{
			char txt[100];
			sprintf(txt,"SUBCASE %d\n",l+1);
			fprintf(pf,txt);
			sprintf(txt,"SUBTITLE = Load case  %d\n",id[l]);
			fprintf(pf,txt);
			sprintf(txt,"LOAD=%d\n",id[l]);
			fprintf(pf,txt);
		}
		delete [] id;

		if(spc1)
		{
			fprintf(pf,"SPC=100\n");
			fprintf(pf,"$\n$\nBEGIN BULK\n");
		} else
		{
			fprintf(pf,"$\n$\nBEGIN BULK\n");
		}

		// correct property type id
		ElementProperties *pep = structdata->elemprop;
		int nelp = structdata->nelp;
		for(int k=0;k<nelp;k++)
		{
			int mtyp = pep[k].config;
			if(mtyp < 1000) continue;
			if(mtyp < 1040) mtyp = 10*(mtyp/10)+1;
			else if(mtyp == 1104) mtyp = 1103;
			else if(mtyp == 1106) mtyp = 1105;
			else if(mtyp > 1110 && mtyp <= 1112) mtyp = 1110;
			else if(mtyp > 1113 && mtyp <= 1115) mtyp = 1113;
			else if(mtyp > 1120 && mtyp <= 1122) mtyp = 1120;
			else if(mtyp > 1123 && mtyp <= 1125) mtyp = 1123;
			else if(mtyp > 1126 && mtyp <= 1128) mtyp = 1126;
			else if(mtyp == 1131) mtyp = 1130;
			pep[k].config = mtyp;
		}
	}
	// Node Data
	fprintf(pf,"$\n$ node data\n$\n");
	writeNodeDataN(spc);
	// constraints as spc1 records
	if(spc1)
	{
		fprintf(pf,"$\n$ constraints\n$\n");
		int id[100];
		int mspc = 0;
		int np = psd->np;
		NodeData *nodedata = psd->nodedata;
		int idd,ida = 0;
		int mid = 0;
		for(int k=0;k<np;k++)
		{
			if(!nodedata[k].pspc) continue;
			idd = nodedata[k].pspc;
			if(idd == ida) continue;
			ida = idd;
			for(int i=0;i<mid;i++)
				if(id[i] == idd) break;
			if(i < mid) continue;
			id[mid++] = idd;
		}
		int *nid = new int[np];
		for(int kid=0;kid<mid;kid++)
		{
			idd = id[kid];
			memset(nid,0,np*sizeof(int));
			for(k=0;k<np;k++)
				if(nodedata[k].pspc == idd) nid[k] = nodedata[k].id;
			int ix2,ix=0;
			while(1)
			{
				while(ix < np && !nid[ix]) ++ix;
				for(ix2=ix+1;ix2<np;ix2++)
					if(nid[ix2] != nid[ix2-1]+1) break;
				if(ix2 > ix+2) // range
				{
					fprintf(pf,"SPC1, 100, %d, %d, THRU, %d\n",idd,nid[ix],nid[ix2-1]);
					for(int i=ix;i<ix2;i++) nid[i] = 0;
				}
				ix = ix2;
				if(ix > np-2) break;
			}
			int kf=0,m=0;
			char xx[9],txt[80];
			sprintf(txt,"SPC1, 100, %d",idd);
			for(ix=0;ix<np;ix++)
			{
				if(!nid[ix]) continue;
				sprintf(xx,",%d",nid[ix]);
				strcat(txt,xx);
				++m;
				if((!kf && m == 6) || (kf && m == 8))
				{
					strcat(txt,"\n");
					fprintf(pf,txt);
					m = 0;
					kf = 1;
					strcpy(txt,"   ,");
				}
			}
			if( m > 0)
			{
				strcat(txt,"\n");
				fprintf(pf,txt);
			}
		}
	}

	// Element Data
	fprintf(pf,"$\n$ element data\n$\n");
	writeElementDataN();
	// coordinate frames
	fprintf(pf,"$\n$ coordinate frames\n$\n");
	writeCoordinateFramesN();
	if(!geom)
	{
		if(mfile) // different file for properties and loads
		{
			fclose(pf);
			char *p = strrchr(file,'.');
			*p = '\0';
			strcat(file,".nprop");
			pf = fopen(file,"w");
		}
		// Material properties
		fprintf(pf,"$\n$ material properties\n$\n");
		writeMaterialPropertiesN();
		// Element Properties
		fprintf(pf,"$\n$ element properties\n$\n");
		writeElementPropertiesN();
		// Distributed loads
		if(mfile) // different file title
		{
			fclose(pf);
			char *p = strrchr(file,'.');
			*p = '\0';
			strcat(file,".nload");
			pf = fopen(file,"w");
		}
		fprintf(pf,"$\n$ distributed loads\n$\n");
		writeDistibutedLoadsN();
		// Node forces
		fprintf(pf,"$\n$ node forces\n$\n");
		writeNodeForcesN();
	}
	if(!geom && !mfile) fprintf(pf,"$\nENDDATA");
	fclose(pf);
	return 0;
}
// Node data
void writeNodeDataN(int spc)
{
	int np = structdata->np;
	NodeData *p = structdata->nodedata;

	for(int k=0;k<np;k++)
	{
		fprintf(pf,"GRID, %d, 0,",p[k].id);
		write_record(p[k].xyz,3,0,0);
		if(geomonly)
		{
			fprintf(pf,"\n");
			continue;
		}
		if(spc && p[k].pspc) 
			fprintf(pf," %d, %d,\n",p[k].cid,p[k].pspc);
		else
			fprintf(pf,"\n");
	}
}
// Element data
void writeElementDataN()
{
	ElementData *pe = structdata->elemdata;
	ElementProperties *pep = structdata->elemprop;
	NodeData *nd = structdata->nodedata;
	int ne = structdata->ne;
	int nelp = structdata->nelp;
	int shape,nodes,k,ie;

	// indices of property records
	if(!geomonly)
	{
		ixprop = new int[ne];
		ixprop2 = new int[ne];
		memset(ixprop,0,ne*sizeof(int));
		memset(ixprop2,0,ne*sizeof(int));
		for(k=0;k<nelp;k++)
		{
			int pid = pep[k].id;
			int config = pep[k].config;
			for(ie=0;ie<ne;ie++)
			{
				int mtype = pe[ie].mechtype;
				if(config > 1000)
				{
					if(ixprop[ie]) continue;
					if(mtype < 40) mtype = 10*(mtype/10)+1;
					if(mtype == 104) mtype = 103;
					if(mtype == 106) mtype = 105;
					if(mtype > 110 && mtype <= 112) mtype = 110;
					if(mtype > 113 && mtype <= 115) mtype= 113;
					if(mtype > 120 && mtype <= 122) mtype = 120;
					if(mtype > 123 && mtype <= 125) mtype = 123;
					if(mtype > 126 && mtype <= 128) mtype = 126;
					if(mtype == 131) mtype = 130;

					if(config < 1040) config = 1000 + 10*(mtype/10)+1;
					if(config == 1104) config = 1103;
					if(config == 1106) config = 1105;
					if(config > 1110 && config <= 1112) config = 1110;
					if(config > 1113 && config <= 1115) config= 1113;
					if(config > 1120 && config <= 1122) config = 1120;
					if(config > 1123 && config <= 1125) config = 1123;
					if(config > 1126 && config <= 1128) config = 1126;
					if(config == 1131) config = 1130;

					if(config != mtype+1000) continue;
					if(pe[ie].pid == pid) ixprop[ie] = k+1;
				} else
				{
					if(ixprop2[ie]) continue;
					if(config != mtype) continue;
					if(pe[ie].pid2 == pid) ixprop2[ie] = k+1;
				}
			}
		}
	} else
	{
		ixprop = NULL;
		ixprop2 = NULL;
	}
	for(ie=0;ie<ne;ie++)
	{
		shape = pe[ie].shape;
		nodes = pe[ie].nodes;
		for(int l=0;l<nodes;l++) // node index to node id
		{
			int ix = pe[ie].lnodes[l];
			if(ix) pe[ie].lnodes[l] = nd[ix-1].id;
		}
		if(shape == 4 || shape == 3) // surface element
		{
			writeSurfaceElement(&pe[ie],ie);
		} else if(shape >= 5 && shape <= 8) // solid element
		{
			writeSolidElement(&pe[ie],ie);
		} else if(shape <= 2) // bar element, point element
		{
			writeBarElement(&pe[ie],ie);
		}
	}
	delete [] ixprop;
	delete [] ixprop2;
	ixprop = NULL;
	ixprop2 = NULL;
}
void writeSurfaceElement(ElementData *pe,int ie)
{
	ElementProperties *pep = structdata->elemprop;
	char *etyp1[] = {"CTRIA3","CTRIAR"};
	char *etyp2[] = {"CQUAD4","CQUADR"};
	char *etyp3[] = {"CQUAD","CQUADX","CTRIAX"};

	int ixetyp = 0;
	int mtyp = pe->mechtype;
	int nodes = pe->nodes;
	int ixp=0,ixp2=0;
	if(!geomonly)
	{
		ixp = ixprop[ie];
		ixp2 = ixprop2[ie];
	}
	int pid=0;
	if(ixp) pid = pep[ixp-1].id;

	if((mtyp%10 == 1 || mtyp%10 == 2) && nodes == 3) // CTRIA3,CTRIAR
	{
		fprintf(pf,"%s, %d, %d,",etyp1[mtyp%10-1],pe->id,pid);
		for(int i=0;i<3;i++)
			fprintf(pf," %d,",pe->lnodes[i]);
		int kzr = 0;
		if(mtyp == 22) kzr = 1;
		if(ixp2)
			writeSurfaceElement2(ixp2,3,kzr);
		else
			fprintf(pf,"\n");
	} else if((mtyp%10 == 1 || mtyp%10 == 2) && nodes == 4) // CQUAD4,CQUADR
	{
		fprintf(pf,"%s, %d, %d,",etyp2[mtyp%10-1],pe->id,pid);
		for(int i=0;i<4;i++)
			fprintf(pf," %d,",pe->lnodes[i]);
		int kzr = 0;
		if(mtyp == 22) kzr = 1;
		if(ixp2)
			writeSurfaceElement2(ixp2,4,kzr);
		else
			fprintf(pf,"\n");
	} else if(mtyp%10 == 1 && nodes == 6) // CTRIA6
	{
		fprintf(pf,"CTRIA6, %d, %d,",pe->id,pid);
		for(int i=0;i<6;i++)
			fprintf(pf," %d,",pe->lnodes[i]);
			fprintf(pf,"\n");
		if(ixp2)
			writeSurfaceElement3(ixp2,5);
		else
			fprintf(pf,"     ,   ,\n");
	} else if(mtyp%10 == 1 && nodes == 8) // CQUAD8
	{
		fprintf(pf,"CQUAD8, %d, %d,",pe->id,pid);
		for(int i=0;i<6;i++)
			fprintf(pf," %d,",pe->lnodes[i]);
			fprintf(pf,"\n");
		fprintf(pf,"     ,");
		for(i=6;i<8;i++)
			fprintf(pf," %d,",pe->lnodes[i]);
		if(ixp2)
			writeSurfaceElement4(ixp2,4);
		else
			fprintf(pf,"\n");
	} else if(mtyp%10 == 3 || mtyp%10 == 4) // CQUAD,CQUADX,CTRIAX
	{
		int i = 0;
		if(mtyp%10 == 4) i = 1;
		if(i == 1 && (nodes == 3 || nodes == 6)) i = 2;
		fprintf(pf,"%s, %d, %d,",etyp3[i],pe->id,pid);
		int m = nodes;
		if(m > 6) m = 6;
		for(i=0;i<m;i++)
			fprintf(pf," %d,",pe->lnodes[i]);
		fprintf(pf,"\n");
		if(nodes > 6)
		{
			fprintf(pf,"     ,");
			for(int i=6;i<nodes;i++)
				fprintf(pf," %d,",pe->lnodes[i]);
			fprintf(pf,"\n");
		}
	} else if(mtyp%10 == 5 && (nodes == 3 || nodes == 6)) // CTRIAX6
	{
		int mid = pep[ixp2-1].matid;
		fprintf(pf,"CTRIAX6, %d, %d,",pe->id,mid);
		for(int i=0;i<nodes;i++)
			fprintf(pf," %d,",pe->lnodes[i]);
		fprintf(pf,"\n");
		if(pep[ixp2-1].nv > 0) write_record(pep[ixp2-1].data,1,1,1);
	} else if(mtyp/10 == 4 && nodes == 4) // CSHREAR
	{
		int mid = pep[ixp2-1].matid;
		fprintf(pf,"CSHEAR, %d, %d,",pe->id,mid);
		for(int i=0;i<4;i++)
			fprintf(pf," %d,",pe->lnodes[i]);
		fprintf(pf,"\n");
	}
}
void writeSurfaceElement2(int ixp2,int n,int kzr)
{
	ElementProperties *pep = structdata->elemprop;
	float *pd = pep[ixp2-1].data;
	float data[10];
	int nw = pep[ixp2-1].nv;
	if(nw < n) n = nw;
	if(kzr) pd[1] = 0.f;

	if((pd[0] == 0.f || pd[0] == -99.f) && (pd[1] == 0.f || pd[1] == -99.f))
		fprintf(pf,"\n");
	else
		write_record(pd,2,1,0);
	if(nw > 2)
	{
		data[0] = data[1] = -99.f;
		memcpy(&data[2],&pd[2],n*sizeof(float));
		write_record(data,n+2,1,1);
	}
}
void writeSurfaceElement3(int ixp2,int n)
{
	ElementProperties *pep = structdata->elemprop;
	float *pd = pep[ixp2-1].data;
	int nv = pep[ixp2-1].nv;

	if(n > nv) n = nv;
	write_record(pd,n,1,1);
}
void writeSurfaceElement4(int ixp2,int n)
{
	ElementProperties *pep = structdata->elemprop;
	float *pd = pep[ixp2-1].data;
	float data[10] = {-99.f,-99.f,-99.f,-99.f,-99.f,-99.f};
	int nw = pep[ixp2-1].nv;
	int kt = 1;

	if((pd[0] == 0.f || pd[0] == -99.f) && (pd[1] == 0.f || pd[1] == -99.f)) kt = 0;
	if(nw <= 2)
	{
		memcpy(&data[n],pd,2*sizeof(float));
	} else
	{
		memcpy(&data[n],pd,2*sizeof(float));
		memcpy(data,&pd[2],n*sizeof(float));
	}
	write_record(data,n+2,1,0);
}
void writeSolidElement(ElementData *pe,int ie)
{
	char *etyp[] = {"CPENTA","CTETRA","CHEXA"};
	float ww[40];
	int shape = pe->shape;
	if(shape != 5 && shape != 7 && shape != 8) return;
	int ix = 2;
	int nodes = pe->nodes;
	if(shape == 7)
		ix = 0;
	else if(shape == 5)
		ix = 1;
	for(int i=0;i<nodes;i++) ww[i] = (float) pe->lnodes[i];
	if(pe->pid == 0) pe->pid = 1;
	fprintf(pf,"%s, %d, %d, ",etyp[ix],pe->id,pe->pid);
	int m = 6;
	if(m > nodes) m = nodes;
	write_recordi(ww,m,1,0);
	nodes -= m;
	ix = m;
	while(nodes > 0)
	{
		m = 8;
		if(m > nodes) m = nodes;
		write_recordi(&ww[ix],m,1,1);
		nodes -= m;
		ix += m;
	}
}
void writeBarElement(ElementData *pe,int ie)
{
	ElementProperties *pep = structdata->elemprop;
	char kkw[10];
	float ww[24];
	int ixp=0,ixp2=0,mtyp = pe->mechtype;

	if(mtyp < 100 || mtyp > 211) return;

	if(mtyp >= 110 && mtyp < 130) // CBAR,CBEAM
	{
		writeCBarElement(pe,ie);
		return;
	}

	int id = pe->id;
	int k1 = pe->lnodes[0];
	int k2 = pe->lnodes[1];
	int k3 = pe->lnodes[2];
	int k4 = pe->lnodes[3];
	int k5 = pe->lnodes[4];

	if(!geomonly)
	{
		ixp = ixprop[ie];
		ixp2 = ixprop2[ie];
	}
	int pid=0;
	if(ixp) pid = pep[ixp-1].id;
	ww[0] = (float) id;
	ww[1] = (float) pid;
	ww[2] = (float) k1;
	ww[3] = (float) k2;
	for(int i=4;i<24;i++) ww[i] = -99.f;
	int mi=0,mr=0;

	if(mtyp == 101) // CONROD
	{
		strcpy(kkw,"CONROD");
		ww[1] = (float) k1;
		ww[2] = (float) k2;
		ww[3] = -99.f;
		mi = 4;
		if(ixp2)
		{
			ww[3] = (float) pep[ixp2-1].matid;
			for(i=0;i<4;i++)
				ww[4+i] = (float) pep[ixp2-1].data[i];
			mr = 4;
		}
	} else if(mtyp == 102) // CROD
	{
		mi = 4;
		strcpy(kkw,"CROD");
	} else if(mtyp == 103 || mtyp == 105) // CELAS1,CDAMP1
	{
		mi = 6;
		strcpy(kkw,"CELAS1");
		if(mtyp == 105) strcpy(kkw,"CDAMP1");
		ww[3] = -99.f;
		ww[4] = (float) k2;
		if(ixp2)
		{
			ww[3] = pep[ixp2-1].data[0];
			ww[5] = pep[ixp2-1].data[1];
		}
		if(!k2) ww[5] = -99.f;
	} else if(mtyp == 104) // CELAS2
	{
		fprintf(pf,"CELAS2, %d,",id);
		ww[3] = -99.f;
		ww[4] = (float) k2;
		if(ixp2)
		{
			ww[1] = pep[ixp2-1].data[0];
			ww[3] = pep[ixp2-1].data[1];
			ww[5] = pep[ixp2-1].data[2];
			ww[6] = pep[ixp2-1].data[3];
			ww[7] = pep[ixp2-1].data[4];
		}
		if(!k2) ww[5] = -99.f;
		write_record(&ww[1],1,0,0);
		write_recordi(&ww[2],4,0,0);
		write_record(&ww[6],2,1,0);
		return;
	} else if(mtyp == 106) // CDAMP2
	{
		fprintf(pf,"CDAMP2, %d,",id);
		ww[1] = -99.f;
		ww[3] = -99.f;
		ww[4] = (float) k2;
		if(ixp2)
		{
			ww[1] = pep[ixp2-1].data[0];
			ww[3] = pep[ixp2-1].data[1];
			ww[5] = pep[ixp2-1].data[2];
		}
		if(!k2) ww[5] = -99.f;
		write_record(&ww[1],1,0,0);
		write_recordi(&ww[2],4,1,0);
		return;
	} else if(mtyp == 108) // CTUBE
	{
		mi = 4;
		strcpy(kkw,"CTUBE");
	} else if(mtyp == 130) // CGAP
	{
		fprintf(pf,"CGAP, ");
		write_recordi(ww,4,0,0);
		if(ixp2)
		{
			for(i=0;i<4;i++)
				ww[4+i] = pep[ixp2-1].data[i];
			write_record(&ww[4],3,0,0);
			write_recordi(&ww[7],1,1,0);
		} else
			fprintf(pf,"\n");
		return;
	} else if(mtyp == 131) // CGAP
	{
		mi = 5;
		strcpy(kkw,"CGAP");
		ww[4] = (float) k3;
	} else if(mtyp == 140) // CVISC
	{
		mi = 4;
		strcpy(kkw,"CVISC");
	} else if(mtyp == 150) // RBAR
	{
		mi = 7;
		strcpy(kkw,"RBAR");
		ww[1] = (float) k1;
		ww[2] = (float) k2;
		ww[3] = -99.f;
		if(ixp2)
		{
			for(i=0;i<4;i++)
				ww[3+i] = pep[ixp2-1].data[i];
		}
	} else if(mtyp == 201) // CONM1
	{
		fprintf(pf,"CONM1, ");
		ww[1] = (float) k1;
		ww[2] = ww[3] = -99.f;
		if(ixp2)
		{
			for(i=0;i<22;i++)
				ww[2+i] = pep[ixp2-1].data[i];
		}
		write_recordi(ww,3,0,0);
		write_record(&ww[3],5,1,0);
		write_record(&ww[8],8,1,1);
		write_record(&ww[16],8,1,1);
		return;
	} else if(mtyp == 202) // CONM2
	{
		fprintf(pf,"CONM2, ");
		ww[1] = (float) k1;
		ww[2] = ww[3] = -99.f;
		if(ixp2)
		{
			for(i=0;i<11;i++)
				ww[2+i] = pep[ixp2-1].data[i];
		}
		write_recordi(ww,3,0,0);
		write_record(&ww[3],4,1,0);
		write_record(&ww[7],6,1,1);
		return;
	} else if(mtyp == 210) // MPC
	{
		if(!ixp2) return;
		int nv = pep[ixp2-1].nv;
		int mk = (nv-1)/3;
		int sid = (int) pep[ixp2-1].data[0];
		fprintf(pf,"MPC, %d",sid);
		for(int l=0;l<mk;l++)
		{
			int k = (int) pep[ixp2-1].data[3*l+1];
			int ci = (int) pep[ixp2-1].data[3*l+2];
			fprintf(pf," %d, %d,",k,ci);
			ww[0] = (float) pep[ixp2-1].data[3*l+3];
			int nl = 0;
			int kz = 0;
			if(l && l%2 == 0) nl = 1;
			if(l%2 == 1 || l == mk-1) kz = 1;
			write_record(ww,1,nl,kz);
		}
		return;
	} else if(mtyp == 211) // MPC1
	{
		if(!ixp2) return;
		int mk = pe->nodes;
		int sid = (int) pep[ixp2-1].data[0];
		fprintf(pf,"MPC, %d",sid);
		for(int l=0;l<mk;l++)
		{
			int k = pe->lnodes[l];
			if(!k) continue;
			if(l && l%2 == 0)
			{
				fprintf(pf,"    ,  ,");
			}
			int ci = (int) pep[ixp2-1].data[2*l+1];
			fprintf(pf," %d, %d,",k,ci);
			ww[0] = (float) pep[ixp2-1].data[2*l+2];
			int kz = 0;
			if(l%2 == 1 || l == mk-1) kz = 1;
			write_record(ww,1,kz,0);
		}
		return;
	}
	if(!mi && !mr) return;
	fprintf(pf,"%s, ",kkw);
	write_recordi(ww,mi,!mr,0);
	if(mr)
		write_record(&ww[mi],mr,1,0);
}
void writeCBarElement(ElementData *pe,int ie)
{
	ElementProperties *pep = structdata->elemprop;
	NodeData *pnode = structdata->nodedata;
	int np = structdata->np;
	float ww[20];
	int ixp=0,ixp2=0;
	int mtyp = pe->mechtype;
	int kzi[20];
	memset(kzi,0,20*sizeof(int));

	int id = pe->id;
	int k1 = pe->lnodes[0];
	int k2 = pe->lnodes[1];
	int k3 = pe->lnodes[2];
	int k4 = pe->lnodes[3];
	int k5 = pe->lnodes[4];

	if(!geomonly)
	{
		ixp = ixprop[ie];
		ixp2 = ixprop2[ie];
	}
	int pid=0;
	if(ixp) pid = pep[ixp-1].id;
	for(int i=0;i<16;i++) ww[i] = -99.f;
	ww[0] = (float) id;
	ww[1] = (float) pid;
	ww[2] = (float) k1;
	ww[3] = (float) k2;
	int m = 4;

	int n = 14;
	if(mtyp/10 == 11) n = 12;
	if(mtyp%10 == 2 || mtyp%10 == 5 || mtyp%10 == 8) n = 2;
	for(i=0;i<4;i++) kzi[i] = 1;
	int nv = pep[ixp2-1].nv;
	if(n > 2)
	{
		for(i=0;i<nv;i++)
			ww[4+i] = pep[ixp2-1].data[i];
	} else
	{
		int ix0 = 16;
		if(mtyp/10 == 11) ix0 = 8;
		for(i=0;i<n;i++)
			ww[ix0+i] = pep[ixp2-1].data[i];
	}
	kzi[8] = kzi[9] = 1;

	if(mtyp < 120)
	{
		fprintf(pf,"CBAR");
		ww[7] = -99.f;
	} else
		fprintf(pf,"CBEAM");

	if(ixp2)
	{
		m = 16;
		if(mtyp/10 == 12) m = 18;
		if(mtyp%10 == 0 || mtyp%10 == 3 || mtyp%10 == 6)
		{
			if(mtyp/10 == 12)
				kzi[7] = 1;
			else
				ww[7] = -99.f;

		} else if(mtyp%10 == 1 || mtyp % 10 == 4 || mtyp%10 == 7)
		{
			ww[4] = (float) k3;
			kzi[4] = 1;
			ww[5] = ww[6] = -99.f;
		} else
		{
			int ixk1=-1,ixk2=-1,ixk4=-1,ixk5=-1;
			ww[4] = (float) k3;
			kzi[4] = 1;
			ww[5] = ww[6] = -99.f;
			for(i=10;i<=15;i++) ww[i] = -99.f;
			if(k1 <= np && pnode[k1-1].id == k1)
			{
				ixk1 = k1-1;
			} else
			{
				for(i=0;i<np;i++)
					if(pnode[i].id == k1) break;
				if(i < np) ixk1 = i;
			}
			if(k2 <= np && pnode[k2-1].id == k2)
			{
				ixk2 = k2-1;
			} else
			{
				for(i=0;i<np;i++)
					if(pnode[i].id == k2) break;
				if(i < np) ixk2 = i;
			}
			if(k4)
			{
				if(k4 <= np && pnode[k4-1].id == k4)
				{
					ixk4 = k4-1;
				} else
				{
					for(i=0;i<np;i++)
						if(pnode[i].id == k4) break;
					if(i < np) ixk4 = 1;
				}
			}
			if(k5)
			{
				if(k5 <= np && pnode[k5-1].id == k5)
				{
					ixk5 = k5-1;
				} else
				{
					for(i=0;i<np;i++)
						if(pnode[i].id == k5) break;
					if(i < np) ixk5 = 1;
				}
			}
			if(ixk1 >= 0 && ixk4 >= 0)
				for(i=0;i<3;i++)
					ww[i+10] = pnode[ixk1].xyz[i] - pnode[ixk4].xyz[i];
			if(ixk2 >= 0 && ixk5 >= 0)
				for(i=0;i<3;i++)
					ww[i+13] = pnode[ixk2].xyz[i] - pnode[ixk5].xyz[i];
		}
	}
	write_recordg(ww,kzi,8,1,0);
	int kz = 0;
	for(i=8;i<m;i++)
		if(ww[i] != 0 && ww[i] != -99.f) kz = 1;
	if(kz && m > 8)
	{
		if(kz) write_recordg(&ww[8],&kzi[8],8,1,1);
	}
	if(kz && mtyp/10 == 12 && m > 16)
		write_recordi(&ww[16],2,1,1);
}
// Material Properties
void writeMaterialPropertiesN()
{
	static char *mtype[] = {"","MAT1","MAT2","MAT3","MAT4","MAT5","","","MAT8",
		"MAT9","MAT10","MATHP"};
	int nmt = sizeof(mtype)/sizeof(mtype[0]);

	int nm = structdata->nm;
	MaterialProperties *pm = structdata->mprop;
	float data[100];
	int mv = 0;

	for(int km=0;km<nm;km++)
	{
		int id = pm[km].id;
		int type = pm[km].type;
		if(type < 1 || type > nmt) continue;
		if(!mtype[type] || !strlen(mtype[type])) continue;

		mv = 0;
		for(int l=0;l<100;l++)
		{
			data[l] = (float) pm[km].data[l];
			if(data[l]) mv = l;
		}
		++mv;
		fprintf(pf,"%s, %d,",mtype[type],id);
		int mcsid = 0,m = mv;
		if(type == 1) // MAT1
		{
			if(mv > 10)
				mcsid = (int) data[10];
			if(m > 7) m = 7;
			write_record(data,m,1,0);
			if(mv > 7) write_record(&data[7],3,!mcsid,1);
			if(mcsid) write_recordi(&data[10],1,1,0);
		} else if(type == 2) // MAT2
		{
			if(mv > 15)
				mcsid = (int) data[15];
			if(m > 7) m = 7;
			write_record(data,m,1,0);
			if(mv > 7)
				write_record(&data[7],8,1,1);
			if(mcsid) write_recordi(&data[15],1,1,1);
		} else if(type <= 10)
		{
			write_recordm(data,mv,7,0);
		} else if(type == 11) // MATHP
		{
			m = mv;
			if(m > 7) m = 7;
			write_record(data,m,1,0);
			data[7] = -99.f;
			for(int i=19;i<23;i++)
				data[i] = -99.f;
			for(i=28;i<31;i++)
				data[i] = -99.f;
			for(i=37;i<39;i++)
				data[i] = -99.f;
			data[46] = -99.f;
			for(i=51;i<54;i++)
				data[i] = -99.f;
			if(mv > 7)
				write_recordi(&data[7],3,1,1);
			int mvv = mv - 15;
			if(mvv > 32) mvv = 32;
			write_recordm(&data[15],mvv,8,1);
			if(mv > 47)
				write_recordi(&data[47],8,1,1);
		}
	}
}

// Element Properties
void writeElementPropertiesN()
{
	static char *type[] = {"ROD","TUBE","L","I","CHAN","T","BOX","BAR",
		"CROSS","H","TI","I1","CHAN1","Z","CHAN2","T2","BOX1","HEXA","HAT"};
	static char *cstr[2] = {"GAUS","GRID"};
	static char *FT[] = {"","HILL","HOFF","TSAI","STRN"};
	static struct
	{
		char name[8];
		int mtyp;
		int nw;
		int mid;
	} prop[] = {
	{"PLPLANE",1011,2,1},
	{"PSHELL",1021,9,1},
	{"PCOMP",1031,100,0},
	{"PSHEAR",1041,4,1},
	{"PLSOLID",1060,1,1},
	{"PSOLID",1061,5,1},
	{"PROD",1102,4,1},
	{"PELAS",1103,3,0},
	{"PDAMP",1105,1,0},
	{"PTUBE",1108,4,1},
	{"PBAR",1110,17,1},
	{"PBARL",1113,100,1},
	{"PBEAM",1120,100,1},
	{"PBCOMP",1123,100,1},
	{"PBEAML",1126,100,1},
	{"PGAP",1130,10,0},
	{"PVISC",1140,2,0},
	{"CBARAO",2110,7,0},
	};

	int nelp = structdata->nelp;
	if(!nelp) return;
	ElementProperties *pep = structdata->elemprop;
	int mprop = sizeof(prop)/sizeof(prop[0]);
	float qq[EPM];
	int kzi[EPM];

	for(int kp=0;kp<nelp;kp++)
	{
		int pid = pep[kp].id;
		int mtyp = pep[kp].config;
		if(mtyp < 1000) continue;
		int nv = pep[kp].nv;
		int mid = pep[kp].matid;

		for(int ip=0;ip<mprop;ip++)
			if(prop[ip].mtyp == mtyp) break;
		if(ip == mprop) continue;

		if(nv > prop[ip].nw) nv = prop[ip].nw;
		if(nv > EPM) nv = EPM;
		memset(kzi,0,EPM*sizeof(int));
		memset(qq,0,EPM*sizeof(float));
		memcpy(qq,pep[kp].data,nv*sizeof(float));

		if(prop[ip].mid) 
			fprintf(pf,"%s, %d, %d,",prop[ip].name,pid,mid);
		else
			fprintf(pf,"%s, %d,",prop[ip].name,pid);

		if(mtyp == 1011) // PLPLANE
		{
			int str = (int) qq[1];
			int cid = (int) qq[0];
			if(str > 1) str = 1;
			if(str < 0) str = 0;
			if(cid == -99) cid = 0;
			if(str > 0 || cid > 0)
				fprintf(pf," %d, %s",cid,cstr[str]);
			fprintf(pf,"\n");

		} else if(mtyp == 1021) // PSHELL
		{
			if(nv == 1)
			{
				for(int k=1;k<9;k++)
					qq[k] = -99.f;
				qq[1] = qq[3] = (float) mid;
				nv = 6;
			}
			kzi[1] = kzi[3] = kzi[8] = 1;
			write_recordg(qq,kzi,6,1,0);
			if(nv > 6) write_recordg(&qq[6],&kzi[6],3,1,1);

		} else if(mtyp == 1031) // PCOMP
		{
			write_record(qq,3,0,0);
			int ft = (int) qq[3];
			if(ft > 0 && ft < 5)
				fprintf(pf," %s, ",FT[ft]);
			else
				fprintf(pf," ,");
			write_record(&qq[4],2,0,0);
			if(qq[6] > 0)
				fprintf(pf," SYM,");
			fprintf(pf,"\n");

			int mw = (nv-8)/4 + 1;
			if(nv <= 7) mw = 0;
			nv -= 7;
			int ix = 7;
			kzi[0] = 1;
			for(int l=0;l<mw;l++)
			{
				write_recordg(&qq[ix],kzi,3,0,!(l%2));
				if(qq[ix+3] > 0 || int(qq[ix+3]) == -88)
					fprintf(pf," YES,");
				else
					fprintf(pf," ,");
				if(l%2 || l == mw-1) fprintf(pf,"\n");
				nv -= 4;
				ix += 4;
			}

		} else if(mtyp == 1041) // PSHEAR
		{
			write_record(qq,4,1,0);

		} else if(mtyp == 1060) // PLSOLID
		{
			int str = (int) pep[kp].data[0];
			if(str > 0)
				fprintf(pf," %s",cstr[1]);
			fprintf(pf,"\n");

		} else if(mtyp == 1061) // PSOLID
		{
			int kzf = (int) qq[4];
			write_recordi(qq,4,!kzf,0);
			if(kzf)
				fprintf(pf," PFLUID, \n");

		} else if(mtyp == 1102) // PROD
		{
			write_record(qq,4,1,0);

		} else if(mtyp == 1103) // PELAS
		{
			write_record(qq,3,1,0);

		} else if(mtyp == 1105) // PDAMP
		{
			write_record(qq,1,1,0);

		} else if(mtyp == 1108) // PTUBE
		{
			write_record(qq,4,1,0);

		} else if(mtyp == 1110) // PBAR
		{
			qq[5] = -99.f;
			write_recordm(qq,nv,6,0);
		} else if(mtyp == 1113) // PBARL
		{
			qq[0] = -99.f;
			int it = (int) qq[1];
			if(it < 0 || it > 18) it = 0;
			fprintf(pf," ,%s,\n",type[it]);
			write_recordm(&qq[6],nv-6,8,1);

		} else if(mtyp == 1120) // PBEAM
		{
			write_recordm(qq,nv,6,0);

		} else if(mtyp == 1123) // PBCOMP
		{
			write_record(qq,6,1,0);
			kzi[6] = 1;
			write_recordg(&qq[6],kzi,7,1,1);
			nv -= 14;
			int ix = 14;
			kzi[3] = 1;
			while(nv > 0)
			{
				write_recordg(&qq[ix],kzi,4,1,1);
				nv -= 8;
				ix += 8;
			}

		} else if(mtyp == 1126) // PBEAML
		{
			qq[0] = -99.f;
			int it = (int) qq[1];
			if(it < 0 || it > 18) it = 0;
			fprintf(pf," ,%s,\n",type[it]);
			write_recordm(&qq[6],nv-6,8,1);
		} else if(mtyp == 1130) // PGAP
		{
			write_recordm(qq,nv,7,0);

		} else if(mtyp == 1140) // PVISC
		{
			write_record(qq,2,1,0);
		} else if(mtyp == 2110) // CBARAO
		{
			int scale = (int) qq[0];
			if(scale <= 1)
				fprintf(pf," ,LE");
			else
				fprintf(pf," ,FR");

			if(scale <= 1 && nv == 4)
			{
				write_recordi(&qq[1],1,0,0);
				write_record(&qq[2],2,1,0);
			} else
				write_record(&qq[1],nv-1,1,0);
		}
	}
	return;
}
// Coordinate Frames
void writeCoordinateFramesN()
{
	static char *nas[] = {"CORD2R","CORD2C","CORD2S"};
	CoorSystem *pc = structdata->coorsys;
	int ncoor = structdata->ncoor;

	for(int k=0;k<ncoor;k++)
	{
		int type = pc[k].ktyp;
		fprintf(pf,"%s, %d, ,",nas[type-1],pc[k].id);
		write_record(pc[k].coors[0],6,1,0);
		write_record(pc[k].coors[2],3,1,1);
	}
}
// Distributed Loads
void writeDistibutedLoadsN()
{
	DistributedLoads *pl = structdata->elemloads;
	int nel = structdata->nel;
	
	for(int kl=0;kl<nel;kl++)
	{
		int ltyp = pl[kl].nastrantyp;
		if(ltyp == 10)
			writeDistibutedLoadsN10(&pl[kl]);
		else if(ltyp == 11)
			writeDistibutedLoadsN11(&pl[kl]);
		else if(ltyp == 12)
			writeDistibutedLoadsN12(&pl[kl]);
		else if(ltyp == 14)
			writeDistibutedLoadsN14(&pl[kl]);
	}
}
void writeDistibutedLoadsN12(DistributedLoads *pl)
{
	float ww[33];
	memset(ww,0,33*sizeof(int));
	for(int ne=1;ne<=32;ne++)
	{
		if(!pl->pdata[ne]) break;
		ww[ne-1] = pl->pdata[ne];
	}
	--ne;
	// zusammenhaengende elementbereiche
	int ix = 1;
	while(1)
	{
		for(int l=ix;l<ne;l++)
			if(int(ww[l]) != int(ww[l-1])+1) break;

		if(l > ix+3)
		{
			fprintf(pf,"PLOAD2, %d,",pl->loadset);
			write_record(&pl->pdata[0],1,0,0);
			fprintf(pf," %d, THRU, %d,\n",int(ww[ix-1]),int(ww[l-1]));
			for(int i=ix-1;i<l;i++) ww[i] = 0.f;
			ix = l+1;
			if(l >= ne-3 || !ww[l]) break;
		} else
		{
			ix = l+1;
			if(l >= ne-3 || !ww[l]) break;
		}
	}

	ix = 0;
	for(int l=0;l<ne;l++)
		if(ww[l]) ww[ix++] = ww[l];
	ne = ix;
	int mw = 6;
	if(ne < 6) mw = ne;
	ix = 0;
	while(ne > 0)
	{
		fprintf(pf,"PLOAD2, %d,",pl->loadset);
		write_record(&pl->pdata[0],1,0,0);
		write_recordi(&ww[ix],mw,1,0);
		ne -= 6;
		ix += 6;
		if(mw > ne) mw = ne;
	}
}
void writeDistibutedLoadsN10(DistributedLoads *pl)
{
	NodeData *nodedata = structdata->nodedata;
	fprintf(pf,"PLOAD, %d,",pl->loadset);
	write_record(pl->pdata,1,0,0);
	float idd[4];
	memset(idd,0,4*sizeof(float));
	for(int i=0;i<4;i++)
	{
		int nid = int(pl->pdata[i+1])-1;
		if(nid >= 0) idd[i] = (float) nodedata[nid].id;
	}
	write_recordi(idd,4,1,0);
}
void writeDistibutedLoadsN14(DistributedLoads *pl)
{
	NodeData *nodedata = structdata->nodedata;
	int ide = pl->id;
	int g1 = (int) pl->pdata[4];
	int g4 = (int) pl->pdata[5];
	if(g1) g1 = nodedata[g1-1].id;
	if(g4) g4 = nodedata[g4-1].id;
	pl->pdata[4] = (float) g1;
	pl->pdata[5] = (float) g4;
	fprintf(pf,"PLOAD4, %d, %d,",pl->loadset,ide);
	write_record(pl->pdata,4,0,0);
	write_recordi(&pl->pdata[4],2,1,0);
	for(int k=9;k>5;k--)
		if(pl->pdata[k]) break;
	if(k > 6)
	{
		int kzi[4];
		memset(kzi,0,4*sizeof(int));
		kzi[0] = 1;
		write_recordg(&pl->pdata[6],kzi,4,1,1);
	}
}
void writeDistibutedLoadsN11(DistributedLoads *pl)
{
	static char *type[] = {"","FX","FY","FZ","FXE","FYE","FZE",
		"MX","MY","MZ","MXE","MYE","MZE"};
	static char *scale[] = {"","LE","FR","LEPR","FRPR"};

	ElementData *pe = structdata->elemdata;
	int ktyp = (int) pl->pdata[0];
	int kscale = (int) pl->pdata[1];
	if(ktyp < 1 || ktyp > 12) ktyp = 1;
	if(kscale < 1 || kscale > 4) kscale = 1;
	int ix = 6;
	while(1)
	{
		int ide = int(pl->pdata[ix++]);
		if(!ide || ix > 32) break;
		ide = pe[ide-1].id;
		fprintf(pf,"PLOAD1, %d, %d, %s, %s,",pl->loadset,ide,type[ktyp],scale[kscale]);
		write_record(&pl->pdata[2],4,1,0);
	}
}
	
// Node Forces
void writeNodeForcesN()
{
	static char *nas[6] = {"FORCE","MOMENT","SPC","SPC","SPCD","SPCD"};
	NodeForces *pfo = structdata->nodeforces;
	int nf = structdata->nf;
	float ww[4],max;

	for(int kf=0;kf<nf;kf++)
	{
		int type = pfo[kf].type;
		if(type < 0 || type > 5) type = 0;
		max = 0.f;
		for(int l=0;l<3;l++)
			if(fabs(pfo[kf].fdata[l]) > max) max = (float)fabs(pfo[kf].fdata[l]);
		if(max < 1.e-8) continue;
		if(type < 2)
		{
			fprintf(pf,"%s, %d, %d, %d,",nas[type],pfo[kf].loadset,
				pfo[kf].id,pfo[kf].cid);
			ww[0] = pfo[kf].scale;
			memcpy(&ww[1],pfo[kf].fdata,3*sizeof(float));
			write_record(ww,4,1,0);
		} else // single-point constraint
		{
			int m = 0;
			for(int l=0;l<3;l++)
			{
				float d = pfo[kf].fdata[l];
				if(!d) continue;
				int com = l+1 + 3*(type-2);
				if(type > 3) com -= 6;
				++m;
				if(m == 3) fprintf(pf,"\n");
				if(m%2 == 1)
					fprintf(pf,"%s, %d, %d, %d,",nas[type],pfo[kf].loadset,pfo[kf].id,com);
				else
					fprintf(pf," %d, %d,",pfo[kf].id,com);
				ww[0] = d;
				write_record(ww,1,0,0);
			}
			fprintf(pf,"\n");
		}
	}
}

void write_record(float w[],int n,int nl,int first)
{
	static char z[] = {'0','1','2','3','4','5','6','7','8','9'};
	int kw,ns;
		
	if(first) fprintf(pf,"     ,");
	for(int i=0;i<n;i++)
	{
		kw = (int) w[i];

		if(kw == -99 || !w[i])
			fprintf(pf," ,");

		else if(kw == -88)
			fprintf(pf," YES,");
		
		else if(kw == -89)
			fprintf(pf," NO,");

		else
		{
			char txt[20];
			float ww = (float) fabs(w[i]);
			if(ww > 99999999.f)
			{
				float www = w[i]/100000000.f;
				sprintf(txt,"%.3f",www);
				int l = strlen(txt);
				if(l > 6) txt[6] = '\0';
				strcat(txt,"+8,");
			} else if(ww > 99999.f)
			{
				float www = w[i]/100000.f;
				sprintf(txt,"%.3f",www);
				int l = strlen(txt);
				if(l > 6) txt[6] = '\0';
				strcat(txt,"+5,");
			} else if(ww < 0.00000001f)
			{
				float www = w[i]*100000000.f;
				sprintf(txt,"%.4f",www);
				int l = strlen(txt);
				if(l > 6) txt[6] = '\0';
				strcat(txt,"-8,");
			} else if(ww < 0.0001f)
			{
				float www = w[i]*100000.f;
				sprintf(txt,"%.4f",www);
				int l = strlen(txt);
				if(l > 6) txt[6] = '\0';
				strcat(txt,"-5,");
			} else
			{
				char form[] = "%.2f";
				ns = 2;
				if(ww < 100.f) ns = 4;
				if(ww < 1.f) ns = 6;
				if(ww >= 1000.f) ns = 1;
				form[2] = z[ns];
				sprintf(txt,form,w[i]);
				int l = strlen(txt);
				if(l > 8) txt[8] = '\0';
				strcat(txt,",");
			}
			fprintf(pf,"%s",txt);
		}
	}
	if(nl) fprintf(pf,"\n");
}
void write_recordi(float w[],int n,int nl,int first)
{
	if(first) fprintf(pf,"     ,");
	for(int i=0;i<n;i++)
	{
		int kw = (int) w[i];

		if(!w[i] || kw == -99)
			fprintf(pf," ,");
		else if(kw == -88)
			fprintf(pf," YES,");
		else if(kw == -89)
			fprintf(pf," NO,");
		else
			fprintf(pf," %d,",kw);
	}
	if(nl) fprintf(pf,"\n");
}
void write_recordm(float w[],int nv,int m1,int first)
{
	if(m1 > nv) m1 = nv;
	write_record(w,m1,1,first);
	int m = 8;
	int ix = m1;
	int nw = nv-m1;
	while(nw > 0)
	{
		if(m > nw) m = nw;
		write_record(&w[ix],m,1,1);
		ix += m;
		nw -= m;
	}
}
void write_recordg(float w[],int kzi[],int n,int nl,int first)
{
	if(first) fprintf(pf,"     ,");
	for(int i=0;i<n;i++)
	{
		int kw = (int) w[i];

		if(!w[i] || kw == -99)
			fprintf(pf," ,");
		else if(kw == -88)
			fprintf(pf," YES,");
		else if(kw == -89)
			fprintf(pf," NO,");
		else if(kzi[i]) // integer
			fprintf(pf," %d,",kw);
		else // real
			write_record(&w[i],1,0,0);
	}
	if(nl) fprintf(pf,"\n");
}